% This script executes a two-level SE and a conventional SE based on the
% same measureent set. The measurements are automatically collected from
% the file produced by measurementGenerator.m. The results are printed and 
% plotted if the variable display is set to true.

%% Input data
load('busbranchmeasurements'); %input file
display = false; %if true, results will be printed and plotted

%% Introducing measurement noise
for s =1:n_sub
    z{s}(z{s}(:,5) > 0.0001^2, 4) = normrnd(z{s}(z{s}(:,5) > 0.0001^2, 4), sqrt(z{s}(z{s}(:,5) > 0.0001^2, 5)));
end

%% Executing local SE 

%Classifying lines:
n_sub = size(z,1); 
i_extbranches = find(linedata(:,1) ~= linedata(:,2)); %substation-external branches 
zcentral = cell(n_sub,1); %pre-processed measurement set for the central level
i_intbranches = cell(1,n_sub); %subsation-internal branches (zero-impedance)
extlinedata = cell(1,n_sub); %array of external branches connected to each substation

if display
    disp('Executing two-level SE')
end

for s = 1:n_sub
    extlinedata{s} = linedata(linedata(:,1) ~= linedata(:,2) & (linedata(:,1) == s | linedata(:,2) == s ),:);
    localbranches = linedata((linedata(:,1) == s & linedata(:,2) == s),3:4);
    zlocal = z{s};
    
    [zcentral{s}, H] = local(zlocal, localbranches, extlinedata{s}, nodetype{s}, s, display);
end

%% Execute central SE and BD processing
[ybus, bsh, g, b] = getGridParameters2(linedata(i_extbranches, [1,2,5,6,7]));

zdata = cat(1, zcentral{[1,2,3,4,5]});
zdata = zdata(~isnan(zdata(:,4)),:);
zdata = sortrows(zdata, [3,1,2]);

%Executing first SE:
[stateestimate, h, r, r_N, J, x] = wlsfunction(zdata, ybus, bsh, g, b);

Del0=stateestimate(:,3); %first estimate
V0=stateestimate(:,2); %first estimate

[r_N_max, i] = max(r_N);

%Bad data processing and rerunning of the SE:
while r_N_max > 3 && J>chi2inv(0.99,size(zdata,1)-n_sub*2+1)
    if display
        disp('Removed bad data in central level SE:')
        disp(zdata(i,:));
    end
    zdata(i,:)=[];
    [stateestimate, h, r, r_N, J, x] = wlsfunction(zdata, ybus, bsh, g, b, x);
    [r_N_max, i] = max(r_N);
end

%Final two-level estimate:
Del=stateestimate(:,3);
V=stateestimate(:,2);

%% Prepare conventional SE
z2 = z; %node-breaker measurement set for the conventional SE
zdata2 = []; %bus-branch measurement set for the conventional SE

for s = 1:n_sub
    
    %Removing extra redundancy from protection measurements and pseudo-meaurements;:
    z2{s}(z2{s}(:,5)>0.010001^2 | z2{s}(:,5)<0.000000001 , :) = [];
    
    %Adding voltage measurement:
    zdata2 = [zdata2; s, 0, 1, mean(z2{s}(z2{s}(:,3)==1, 4)), mean(z2{s}(z2{s}(:,3)==1, 5))/length(z2{s}(z2{s}(:,3)==1, 5))];
    
    %Adding power injections:
    i_inj = find(nodetype{s}(:,2)== 1 |  nodetype{s}(:,2)== 2 | nodetype{s}(:,2)== 3);
    if isempty(i_inj)
        zdata2 = [zdata2;   s, 0, 2, 0, 10^-10;   s, 0, 3, 0, 10^-10];
    end
    if isequal(i_inj, sort(z2{s}(z2{s}(:,3) == 2,1)))
        zdata2 = [zdata2; s, 0, 2, sum(z2{s}(z2{s}(:,3) == 2, 4)), sum(z2{s}(z2{s}(:,3) == 2, 5))];
    end
    if isequal(i_inj, sort(z2{s}(z2{s}(:,3) == 3,1)))
        zdata2 = [zdata2; s, 0, 3, sum(z2{s}(z2{s}(:,3) == 3, 4)), sum(z2{s}(z2{s}(:,3) == 3, 5))];
    end
    
    %Adding power flows:
    i_flow = find(nodetype{s}(:,2)== 5); %indicates bay nodes
    for i = 1:length(i_flow) %active power flow
        i_meas = find(z2{s}(:,3) == 4 &  z2{s}(:,1) == i_flow(i)); %active power flow and belongs to indexed nodes in i_flow
        if i_meas %is not empty
            to_sub = max([extlinedata{s}(  extlinedata{s}(:,1) == s & extlinedata{s}(:,3) == i_flow(i), 2), ...
                        extlinedata{s}(  extlinedata{s}(:,2) == s & extlinedata{s}(:,4) == i_flow(i), 1)]);
            zdata2 = [zdata2; s, to_sub, 4, -z2{s}(i_meas,4), z2{s}(i_meas,5)];
        end
    end
    for i = 1:length(i_flow) %reactive power flow
        i_meas = find(z2{s}(:,3) == 5 &  z2{s}(:,1) == i_flow(i)); %reactive power flow and belongs to indexed nodes in i_flow
        if i_meas %is not empty
            to_sub = max([extlinedata{s}(  extlinedata{s}(:,1) == s & extlinedata{s}(:,3) == i_flow(i), 2), ...
                        extlinedata{s}(  extlinedata{s}(:,2) == s & extlinedata{s}(:,4) == i_flow(i), 1)]);
            zdata2 = [zdata2; s, to_sub, 5, -z2{s}(i_meas,4), z2{s}(i_meas,5)];
        end
    end
end
zdata2 = sortrows(zdata2, [3,1,2]);


%% Executing conventional SE
if display
    disp('Executing conventional SE')
end

%Executing first SE:
[stateestimate, h2, r, r_N2, J, x] = wlsfunction(zdata2, ybus, bsh, g, b);

[r_N_max2, i] = max(r_N2);
 
%Bad data processing and rerunning of the SE:
while r_N_max2 > 3 && J>chi2inv(0.95,size(zdata,1)-n_sub*2+1)
    if display
        disp('Removed bad data in conventional SE:')
        disp(zdata2(i,:));
    end
    zdata2(i,:)=[];
    [stateestimate, h2, r, r_N2, J, x] = wlsfunction(zdata2, ybus, bsh, g, b, x);
    [r_N_max2, i] = max(r_N2);
end

%Final conventional estimate:
Del2=stateestimate(:,3);
V2=stateestimate(:,2);

%% Displaying the outputs

if display
    %% Printing results:
    disp('---------------------------------------------')
    disp('---------------- Load flow ------------------');
    disp('---------------------------------------------');
    disp('| Bus |    V   |  Angle  | ');
    disp('| No  |   pu   |  Degree | ');
    disp('--------------------------');
    for m = 1:n_sub
        fprintf('%4g', m); fprintf('  %8.4f', Vreal(m)); fprintf('   %8.4f', delreal(m)-delreal(1)); fprintf('\n');
    end
    disp('---------------------------------------------');
    disp('----- Conventional State Estimation  --------');
    disp('---------------------------------------------');
    disp('| Bus |    V   |  Angle  | ');
    disp('| No  |   pu   |  Degree | ');
    disp('--------------------------');
    for m = 1:n_sub
        fprintf('%4g', m); fprintf('  %8.4f', V2(m)); fprintf('   %8.4f', Del2(m)); fprintf('\n');
    end
    disp('---------------------------------------------');
    disp('------- Two-level State Estimation  ---------');
    disp('---------------------------------------------');
    disp('| Bus |    V   |  Angle  | ');
    disp('| No  |   pu   |  Degree | ');
    disp('--------------------------');
    for m = 1:n_sub
        fprintf('%4g', m); fprintf('  %8.4f', V(m)); fprintf('   %8.4f', Del(m)); fprintf('\n');
    end
    
    %% Plotting results
    figure(1)
    K = 1:n_sub;
    plot(K, V2, 'b:*',K,V,'r:*',K,Vreal,'b--o', 'linewidth',1.5)
    title('Voltage magnitude estimate')
    xlabel('Bus number')
    xlim([1 n_sub])
    ylabel('Voltage (p.u.)')
    legend('Conventional estimate', 'Two-level estimate','True Value')
    grid on
    
    figure(2)
    j = 1:n_sub;
    plot(j, Del2, 'b:*', j,Del,'r:*',j,delreal,'b--o', 'linewidth',1.5)
    title('Voltage phase angle estimate')
    xlabel('Bus number')
    xlim([1 n_sub])
    ylabel('Voltage phase angle (degrees)')
    legend('Conventional estimate', 'Two-level estimate','True Value')
    grid on
end