% This script collects the inputs from excel files, executes a load, and
% prepares the measurements on the node-breaker level. It also provides
% linedata and true state. The output is stored in a file, so that this only
% has to be run every time the inputs change.

clear; clc;

%% Setting testing parameters and data source
toperror = false; %if true, a bus split topology error is introduced in substation 2.
measurementdatabase = 'database/meas_database_case1.xlsx';


%% Set measurement parameters
rii_zeroinj = 0.00000001^2; %10^-20;
rii_CBvolt = 0.0000001^2; %10^-20;
rii_CBangl = 0.0000001^2; %10^-20;
rii_V = 0.005^2;
rii_Pinj = 0.01^2;
rii_Qinj = 0.01^2;
rii_P = 0.01^2;
rii_Q = 0.01^2;
rii_CB_P = 0.0000001^2; %10^-20;
rii_CB_Q = 0.0000001^2; %10^-20;

MVAbase = 100;

%% Preparing cd.branch (Matpower input)
cd.version = '2';
cd.baseMVA = MVAbase;

data.lines = xlsread('database/lines_database.xlsx');

n_sub = max(max(data.lines(:,1:2)));
a = size(data.lines,1);
cd.branch = [data.lines(:,3:4), data.lines(:,5:7), zeros(a, 5), ones(a,1), -360*ones(a,1), 360*ones(a,1)];

n_nodes = zeros(n_sub,1);
nodenum = zeros(n_sub,1);
i_nodes = cell(n_sub,1);
n=1;
for s = 1:n_sub
    nodenum(s) = sum(n_nodes);
    n_nodes(s) = max(max(data.lines(data.lines(:,1)==s & data.lines(:,2)==s,3:4)));
    i_nodes{s} = (1+nodenum(s)):(nodenum(s)+n_nodes(s));
end

cd.branch(:,1) = nodenum(data.lines(:,1))+data.lines(:,3);
cd.branch(:,2) = nodenum(data.lines(:,2))+data.lines(:,4);

data.CB = xlsread(measurementdatabase, 'OpenCBs');
b = size(data.CB,1);
openCBs = zeros(b,2);
openCBs(:,1) = nodenum(data.CB(:,1))+data.CB(:,2);
openCBs(:,2) = nodenum(data.CB(:,1))+data.CB(:,3);

for i = 1:b
    cd.branch((cd.branch(:,1) == openCBs(i,1) & cd.branch(:,2) == openCBs(i,2)) | ...
        (cd.branch(:,1) == openCBs(i,2) & cd.branch(:,2) == openCBs(i,1)) ,4) = 10000;
end
cd.branch(cd.branch(:,4)==0, 4) = 0.00001;

%% Preparing cd.bus and cd.gen (Matpower input)
ntot_nodes = sum(n_nodes);
cd.bus = [zeros(ntot_nodes,6), ones(ntot_nodes, 1), ones(ntot_nodes,1), zeros(ntot_nodes, 2), ones(ntot_nodes,1), ...
    1.06*ones(ntot_nodes, 1), 0.94*ones(ntot_nodes, 1)];
cd.gen = [];

for s = 1:n_sub
    data.operation{s} = xlsread('database/operation_database.xlsx', num2str(s));
    nodetype{s} = data.operation{s}(:,1:2);
    
    for i = 1:n_nodes(s)
        n = i_nodes{s}(i);
        cd.bus(n,1) = n;
        cd.bus(n,2) = data.operation{s}(i,2);
        
        if data.operation{s}(i,2) == 1 %load bus
            cd.bus(n,3:4) = -data.operation{s}(i,3:4);
            
        elseif data.operation{s}(i,2) == 4 || data.operation{s}(i,2) == 5 %zero-injection bus
            cd.bus(n,2) = 1;
            cd.bus(n,3:4) = [0,0];
            
        elseif data.operation{s}(i,2) == 2 %generator bus
            gen = [n, data.operation{s}(i,3), 0, 1000, -200, ...
                data.operation{s}(i,5), 0, 1, 1000, 0, zeros(1,11)];
            cd.gen = [cd.gen; gen];
            
        elseif data.operation{s}(i,2) == 3 %slack bus
            gen = [n, 0, 0, 1000, -200, ...
                data.operation{s}(i,5), 0, 1, 1000, 0, zeros(1,11)];
            cd.gen = [cd.gen; gen];
            cd.bus(n,8) = data.operation{s}(i,5);
        end
    end
end

pfres = runpf(cd); %runs Matpower load flow

%% Preparing the measurement set
data.meas = cell(n_sub,1);
z = cell(n_sub,1);
for s=1:n_sub
    
    %Find conventional measurements
    data.meas{s} = xlsread(measurementdatabase, num2str(s));
    
    [n_meas, temp] = size(data.meas{s});
    data.meas{s} = [data.meas{s}, nan(n_meas, 5-temp)];
    
    %Find zero-injection nodes
    zeroinj_nodes = data.operation{s}(data.operation{s}(:,2) == 4,1);
    n_zeroinj = length(zeroinj_nodes);
    
    %Find CB signals
    lines = data.lines(data.lines(:,1) == s & data.lines(:,2) == s,3:4);
    openCBs = data.CB(data.CB(:,1)==s,2:3);
    n_openCBs = size(openCBs,1);
    n_closedCBs = size(lines,1)-n_openCBs;
    closedCBs = zeros(n_closedCBs,2);
    count = 1;
    for i = 1:size(lines,1)
        if max(sum((lines(i,:) == openCBs),2)) ~=2
            closedCBs(count,:) = lines(i,:);
            count = count + 1;
        end
    end
    
    if s == 2 && toperror
        openCBs = [openCBs; 4,5];
        n_openCBs = n_openCBs + 1;
        closedCBs(4,:)=[];
        n_closedCBs = n_closedCBs - 1;
    end
    
    %Pre-allocating measurement matrix
    z{s} = zeros(n_meas, 5); 
    
    %Conventional
    for m = 1:n_meas
        meas = data.meas{s}(m,:);
        node = i_nodes{s}(meas(1));
        if meas(3) == 1
            z{s}(m,:) = [meas(1),  0,  1,   pfres.bus(pfres.bus(:,1)==node, 8),   rii_V]; 
        elseif meas(3) == 2
            if cd.bus(node,2) == 1
                z{s}(m,:) = [meas(1),  0,  2,   -pfres.bus(pfres.bus(:,1)==node, 3)/MVAbase,   rii_Pinj]; 
            else
                z{s}(m,:) = [meas(1),  0,  2,   pfres.gen(pfres.gen(:,1)==node, 2)/MVAbase,   rii_Pinj]; 
            end
        elseif meas(3) == 3
            if cd.bus(node,2) == 1
                z{s}(m,:) = [meas(1),  0,  3,   -pfres.bus(pfres.bus(:,1)==node, 4)/MVAbase,   rii_Qinj]; 
            else
                z{s}(m,:) = [meas(1),  0,  3,   pfres.gen(pfres.gen(:,1)==node, 3)/MVAbase,   rii_Qinj]; 
            end
        elseif meas(3) == 4
            if isnan(meas(2)) || meas(2) == 0 %if node injection is from incoming line
                index1 = find(data.lines(:,1) == s & data.lines(:,2)~=s & data.lines(:,3) == meas(1));
                index2 = find(data.lines(:,2) == s & data.lines(:,1)~=s & data.lines(:,4) == meas(1));
                if isempty(index2)
                    t_node = i_nodes{data.lines(index1,2)}(data.lines(index1,4));
                else
                    t_node = i_nodes{data.lines(index2,1)}(data.lines(index2,3));
                end
                if sum(pfres.branch(:,1)==node & pfres.branch(:,2)==t_node)==0
                    z{s}(m,:) = [meas(1),  0,  4, -pfres.branch(pfres.branch(:,1)==t_node & pfres.branch(:,2)==node, 16)/MVAbase,   rii_P];
                else
                    z{s}(m,:) = [meas(1),  0,  4,  -pfres.branch(pfres.branch(:,1)==node & pfres.branch(:,2)==t_node, 14)/MVAbase,   rii_P];
                end
            else
                t_node = i_nodes{s}(meas(2));
                if sum(pfres.branch(:,1) == t_node & pfres.branch(:,2)==node) == 0
                    z{s}(m,:) = [meas(1),  meas(2),  4,   pfres.branch(pfres.branch(:,1)==node & pfres.branch(:,2)==t_node, 14)/MVAbase,   rii_P];
                else
                    z{s}(m,:) = [meas(1),  meas(2),  4,   pfres.branch(pfres.branch(:,1) == t_node & pfres.branch(:,2)==node, 16)/MVAbase,   rii_P];
                end
                
            end
        elseif meas(3) == 5
            if isnan(meas(2)) || meas(2) == 0 %if node injection is from incoming line
                index1 = find(data.lines(:,1) == s & data.lines(:,2)~=s & data.lines(:,3) == meas(1));
                index2 = find(data.lines(:,2) == s & data.lines(:,1)~=s & data.lines(:,4) == meas(1));
                if isempty(index2)
                    t_node = i_nodes{data.lines(index1,2)}(data.lines(index1,4));
                else
                    t_node = i_nodes{data.lines(index2,1)}(data.lines(index2,3));
                end
                if sum(pfres.branch(:,1)==node & pfres.branch(:,2)==t_node)==0
                    z{s}(m,:) = [meas(1),  0,  5, -pfres.branch(pfres.branch(:,1)==t_node & pfres.branch(:,2)==node, 17)/MVAbase,   rii_Q];
                else
                    z{s}(m,:) = [meas(1),  0,  5,  -pfres.branch(pfres.branch(:,1)==node & pfres.branch(:,2)==t_node, 15)/MVAbase,   rii_Q];
                end
            else
                t_node = i_nodes{s}(meas(2));
                if sum(pfres.branch(:,1) == t_node & pfres.branch(:,2)==node) == 0
                    z{s}(m,:) = [meas(1),  meas(2),  5,   pfres.branch(pfres.branch(:,1)==node & pfres.branch(:,2)==t_node, 15)/MVAbase,   rii_Q];
                else
                    z{s}(m,:) = [meas(1),  meas(2),  5,   pfres.branch(pfres.branch(:,1) == t_node & pfres.branch(:,2)==node, 17)/MVAbase,   rii_Q];
                end
                
            end
        end
        if ~isnan(meas(4)) %if a value is given manually in the excel file
         	z{s}(m,4) = meas(4);
        end
        if ~isnan(meas(5)) %if a variance is given manually in the excel file
         	z{s}(m,5) = meas(5)^2;
        end
    end
    
    %Preparing zero-injection virtual measurements        
    z{s} = [z{s};
        zeroinj_nodes, zeros(n_zeroinj,1), 2*ones(n_zeroinj,1), ...
        zeros(n_zeroinj,1), rii_zeroinj*ones(n_zeroinj,1)];
    z{s} = [z{s};
        zeroinj_nodes, zeros(n_zeroinj,1), 3*ones(n_zeroinj,1), ...
        zeros(n_zeroinj,1), rii_zeroinj*ones(n_zeroinj,1)];
    
    %Preparing closed CB virtual measurements
    z{s} = [z{s};
        closedCBs(:,1), closedCBs(:,2), 10*ones(n_closedCBs,1), ...
        zeros(n_closedCBs,1), rii_CBangl*ones(n_closedCBs,1)];
    z{s} = [z{s}; 
        closedCBs(:,1), closedCBs(:,2), 11*ones(n_closedCBs,1), ...
        zeros(n_closedCBs,1), rii_CBvolt*ones(n_closedCBs,1)];
    
    %Preparing open CB virtual measurements
    z{s} = [z{s};
        openCBs(:,1), openCBs(:,2), 12*ones(n_openCBs,1), ...
        zeros(n_openCBs,1), rii_CB_P*ones(n_openCBs,1)];
    z{s} = [z{s};
        openCBs(:,1), openCBs(:,2), 13*ones(n_openCBs,1), ...
        zeros(n_openCBs,1), rii_CB_Q*ones(n_openCBs,1)];
end


%% Prepare linedata
linedata = data.lines;
Vreal = pfres.bus(nodenum+1,8);
delreal = pfres.bus(nodenum+1,9);

%% Save linedata
save('busbranchmeasurements', 'linedata', 'z', 'Vreal', 'delreal', 'nodetype');
save('realestimate', 'Vreal', 'delreal')
